/*
 ============================================================================
 Name        : lex_make.c
 Author      : H. Carl Ott
 Version     : see below
 Copyright   : August-2015
 Description : Read config text file, process text files and generate .bin file for RPTR-NEX
 ============================================================================
 */


#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <stdint.h>


#define	VERSION			"LEXICON MAKE UTIL 1.4, August 2015\n"

#define MAX_WORD_FILES 8
#define CONFIG_LINES (3 + MAX_WORD_FILES)

#define	LEX_CFG_OFFSET		0x7000			// fixed offset to cfg struct in the output file
#define	LEX_WORDS_OFFSET	0x7040			// fixed offset to word storage in the output file
#define MAX_FILE_SIZE		0x20000			// max file size
#define	WORD_BUFFER_SIZE	150000			// ram allocated for slw and flw buffers
#define WORD_DELIMS			" ,."			// valid word delims


/**
 * Prototypes
 */

int	process_file(FILE * file);


/**
 * lexicon config struct,
 * shared with rptr-nex app
 */

typedef struct{
	int16_t		lex_sig;		// valid sig (not used)
	uint8_t  	name[6];		// lexicon name
	uint32_t	flw_offset;		// FLW data offset
	uint16_t	flw_cnt;		// number of FLW
	uint32_t	slw_offset;   	// SLW data offset
	uint16_t	slw_cnt;		// number of SLW
}lex_cfg_t;



/**
 * globals
 */
int 	line_cnt = 0;								// stats
int		slw_cnt	 = 0;								//
int		flw_cnt	= 0;								//
int		slw_dirty_cnt = 0;						    // stats
int		flw_dirty_cnt = 0;							// stats

char	slw_buff[WORD_BUFFER_SIZE];					// buffer for all packed six letter words
char	flw_buff[WORD_BUFFER_SIZE];					// buffer for all packed four letter words

char	*pslw;									    // index for SLWs
char	*pflw;									    // index for FLWs

lex_cfg_t	lex_cfg;								// lexicon config struct,  let RPTR-NEX app know where data is located


/**
 * main
 */

int main(int argc, char *argv[]) {
	char  	line_buff[255];								// gp line buffer

	int		i = 0;										// gp counter
	char	cfg_line[CONFIG_LINES][255];			    // config file text file copied to array
	int		byte;										// gp
	int		ret_val;									// gp

	memset(line_buff,0,sizeof(line_buff));				// init line buffer
	memset(cfg_line, 0, sizeof(cfg_line));				// init config array

	FILE	*cfg_file;									// file handles
	FILE	*inp_bin_file;
	FILE	*out_bin_file;
	FILE	*wrd_file;

	fpos_t		file_size;

	printf(VERSION);

	/* check for proper number of cmd line args */
	if (argc != 2) {
		printf("ERROR: Wrong number of arguments. Usage \"%s config_file\"\n", argv[0]);
		exit(EXIT_FAILURE);
	}

	/* try to open config file */
	cfg_file = fopen(argv[1], "r");

	if (cfg_file == NULL) {
		printf("ERROR: Can't open config file:\"%s\"\n", argv[1]);
		exit(EXIT_FAILURE);
	}

	while ( fgets(line_buff, sizeof(line_buff), cfg_file) != NULL)	{		// read line from file
		line_buff[strcspn(line_buff, "\n" )] = '\0';						// convert the newline to null
		line_buff[strcspn(line_buff, "\r" )] = '\0';						// convert the newline to null

		if (i < CONFIG_LINES) {
			strcpy(cfg_line[i++], line_buff);									// copy to config array
		}

		else{
			printf("WARNING: Config line %d discarded.\n", i+1);
			i++;
		}


	}

	if (cfg_file){
		fclose(cfg_file);	// done with cfg
	}

	pslw = slw_buff;		// init indexes to start of packed buffers
	pflw = flw_buff;

	for (i = 3; i < CONFIG_LINES; ++i) {	// process word source files

		if( *cfg_line[i] != 0x0){			// make sure line is not empty

			wrd_file = 	fopen(cfg_line[i], "r");	// open file

			if (wrd_file){						 	// see if opened ok
				ret_val = process_file(wrd_file);	// process the file
				printf("Processed \"%s\", %d total words were processed from text file.\n",cfg_line[i], ret_val);
				fclose(wrd_file);

			}

			else{
				printf("ERROR: Can't open \"%s\" for processing.\n",cfg_line[i]);		// error on file open
			}



		}

	}


	/* print some stats to console */
	printf("Total lines processed = %d\n",line_cnt);
	printf("Total regular four letter words = %d\n",flw_cnt);
	printf("Total regular six letter words = %d\n",slw_cnt);
	printf("Total four letter dirty words = %d\n",flw_dirty_cnt);
	printf("Total six letter dirty words = %d\n",slw_dirty_cnt);
	printf("Total words = %d\n",slw_dirty_cnt + flw_dirty_cnt + flw_cnt + slw_cnt);

	printf("Now generating output bin file.\n");

	/* get source .bin code */
	inp_bin_file = fopen(cfg_line[0], "rb");

	if (inp_bin_file == NULL) {
		fprintf(stderr, "ERROR, Can't open RPTR inp bin file:\"%s\".\n", cfg_line[0]);
		exit(EXIT_FAILURE);
	}


	/* open output file, binary mode */
	out_bin_file = fopen(cfg_line[1], "wb");

	if (out_bin_file == NULL) {
		fprintf(stderr, "ERROR: Can't open output bin file:\"%s\" for writing.\n", cfg_line[1]);
		exit(EXIT_FAILURE);
	}


	/* just copy input file  to output file  */
	while ( (byte = fgetc(inp_bin_file) ) != EOF){
		fputc(byte, out_bin_file);
	}

	if (inp_bin_file){
		fclose(inp_bin_file);									// done with input .bin file
	}

	fseek(out_bin_file, LEX_CFG_OFFSET, SEEK_SET );				// reset file pointer to cfg struct offset

	lex_cfg.lex_sig = 0xaa55;									// set valid .sig

	memcpy(lex_cfg.name, cfg_line[2], sizeof(lex_cfg.name));	// copy the lexicon name, todo fix short line garbage

	lex_cfg.slw_cnt = slw_cnt + slw_dirty_cnt;					// get the counts and offsets
	lex_cfg.flw_cnt = flw_cnt + flw_dirty_cnt;
	lex_cfg.slw_offset = LEX_WORDS_OFFSET + 0;					// slw first to get a 4 byte alignment
	lex_cfg.flw_offset = LEX_WORDS_OFFSET + ((slw_cnt + slw_dirty_cnt) * 4);		// flw are right after the flw

	fwrite(&lex_cfg, sizeof(lex_cfg), 1, out_bin_file );		// write the config struct to output file


	fseek(out_bin_file, LEX_WORDS_OFFSET, SEEK_SET );			// now set file pointer to the word storage start location
	fwrite(&slw_buff, 4, slw_cnt + slw_dirty_cnt, out_bin_file );				// SLW data first
	fwrite(&flw_buff, 3, flw_cnt + flw_dirty_cnt, out_bin_file );				// then FLW

	fgetpos(out_bin_file, &file_size);							// get current file write pointer (file size)

	if (out_bin_file){
		if ( (fclose(out_bin_file)) == 0 ) {						// see if file closed ok

			printf("Lexicon \"%s\" output to file:\"%s\".\n",cfg_line[2],cfg_line[1] );

			if (file_size  > MAX_FILE_SIZE){
				printf("WARNING: Output file might be too large. %d bytes.\n", (int)file_size);
			}

			else{
				printf("Looks good to me. %d bytes were written to file.\n", (int)file_size);
			}

		}

		else{
			printf("ERROR: Problem closing output file.\n");
		}
	}

	return EXIT_SUCCESS;
}


/**
 * read word source text file,
 * parse and pack SLW and FLWs into ram buffers, update counts
 * any word that has at least one capitalized letter is flagged as a dirty word.
 */

int	process_file(FILE * file){

	char  	line_buff[255];								// gp line buffer
	char  	word_buff[16];								// word buffer
	char 	*pword;										// word pointer
	char	*pchar;										// char pointer
	int	 	dirty_flag = 0;								// was this a naughty word?
	unsigned int word_packed;							// place to build packed word  (minimum 32 bits)
	int		words = 0;

	if (file == NULL){									// check for valid fp
		return 0;
	}

	memset(line_buff,0,sizeof(line_buff));				// init line buffer

	while ( fgets(line_buff, sizeof(line_buff), file) != NULL)				// grab a full line
	{

		line_buff[strcspn(line_buff, "\n" )] = '\0';						// convert the newline to zero terminated
		line_buff[strcspn(line_buff, "\r" )] = '\0';						// convert the cr to zero terminated
		line_cnt++;															// update line count stats

		if (strncmp ("//", line_buff, 2) == 0){								// if line starts with "//" treat as comment and discard
			continue;														// loop for next line
		}

		pword = strtok(line_buff,WORD_DELIMS);								// tokenize line buff

		while (pword != NULL){												// loop through each word on line

			strcpy(word_buff, pword);										// get the word into work buff
			dirty_flag = 0;													// clear flag
			word_packed = 0;										   		// init packed workspace

			/* process SLW */
			if  (strlen(word_buff) == 6){

				for (pchar = word_buff; *pchar; ++pchar){

					if(isupper(*pchar) ) {									// if any letter is caps, dirty
						dirty_flag = 1;
					}

					word_packed <<= 5;										// make room
					word_packed |= ((*pchar - 0x61) & 0x1f);				// or in new letter
				}

				word_packed <<= 2;											// now shift to make room for the two flag bits

				if(pslw < (slw_buff + WORD_BUFFER_SIZE + 4) ){				// check buffer


					if	(dirty_flag){											// copy dirty flag to lsb
						word_packed |= 0x01;
						slw_dirty_cnt++;										// bump stats
					}

					else{
						slw_cnt++;
					}

					*pslw++ = (word_packed >> 24) & 0xff;						// write to buffer
					*pslw++ = (word_packed >> 16) & 0xff;
					*pslw++ = (word_packed >> 8) & 0xff;
					*pslw++ = word_packed & 0xff;
					words++;
				}


			}

			/* process FLW */
			else if  (strlen(word_buff) == 4){

				for (pchar = word_buff; *pchar;  ++pchar){

					if(isupper(*pchar) ) {							// if any letter is caps, dirty
						dirty_flag = 1;
					}

					word_packed <<= 5;								// make room for new letter
					word_packed |= ((*pchar - 0x61) & 0x1f);		// or in new letter
				}

				word_packed <<= 4;									//  shift for flag bits


				if(pflw < (flw_buff + WORD_BUFFER_SIZE + 3) ){			// check buffer

					if	(dirty_flag){									// copy dirty flag to lsb
						word_packed |= 0x01;
						flw_dirty_cnt++;								// bump stats
					}

					else{
						flw_cnt++;
					}


					*pflw++ = (word_packed >> 16) & 0xff;
					*pflw++ = (word_packed >> 8) & 0xff;
					*pflw++ = word_packed & 0xff;
					words++;
					// bump stats
				}
			}


			/* process other length words */
			else{			// wrong word length, ignore it

			}

			pword = strtok(NULL, WORD_DELIMS);			// advance to next word in line buff

		}


	}


	return words;

}

